unsigned int *copyseq(unsigned int *expansion, int exp_len)
{
  int i;
  unsigned int *copy = malloc(sizeof(int) * exp_len);
  for (i = 0; i < exp_len; i++) copy[i] = expansion[i];
  return(copy);
}

struct NODE {
  unsigned int c; // ncurses keystrokes are ints, not chars.
  unsigned char ends_word; // using char as a BOOL here
  unsigned char ends_sibling_list;  // => sibling == 0
  unsigned char is_macro; // if true, the keystroke sequence this far is
                          // expandable.  However it *may* not be complete
                          // and if there are children, then this expansion
                          // should only be invoked if there is a timeout.
                          // Generally with prefix keys we want to avoid this!
                          // (and in the first implementation, is_macro will
                          //  only be true if child == NULL also.)
  unsigned int expansion_length; unsigned int *expansion;
  struct NODE *sibling;
  struct NODE *child;
};

struct NODE *cellarray;
static int nextfree = 0;

struct NODE *add_sequence(struct NODE *base, int *s, int s_len, unsigned int *expansion, int exp_len)
{
  struct NODE *each;
  struct NODE *cell;
  int c = *s;

  each = base;
  if (keydebug) fprintf(stderr, "add_seq[l=%d]: %d\r\n", s_len, c);
  while (each != NULL) { // loop over siblings to match first character
    if (keydebug) fprintf(stderr, "check sibling[l=%d]: %d against %d\r\n", s_len, c, each->c);
    if (each->c == c) {
      if (s_len == 1) {      // This is the end of a word, a longer version of which exists in the trie
        each->expansion_length = exp_len;
        each->expansion = copyseq(expansion, exp_len);
        each->ends_word = 1;   // so all we need to do is associate the data here..
        each->is_macro = 1;
      } else each->child = add_sequence(each->child, s+1, s_len-1, expansion, exp_len);
      return base;
    }
    if (each->ends_sibling_list == 1/*TRUE*/) { assert(each->sibling == NULL); break; }
    each = each->sibling;
  }
  // first char was not found, so add it to the siblings at this level, then add rest of string.
  if (keydebug) fprintf(stderr, "create child[l=%d]: %d\r\n", s_len, c);
  cell = &cellarray[nextfree++]; //malloc(sizeof(struct NODE));
  cell->c = c;
  cell->sibling = base;  // last in chain is NULL
  cell->ends_sibling_list = ((base == NULL) ? 1/*TRUE*/ : 0/*FALSE*/);
  cell->child = NULL;
  base = cell; // As per Rainers comments, create list in reverse order...
  if (s_len == 1) {
    cell->ends_word = 1/*TRUE*/;
    cell->is_macro = 1/*TRUE*/;
    cell->expansion_length = exp_len;
    cell->expansion = copyseq(expansion, exp_len);
  } else {
    cell->ends_word = 0/*FALSE*/;
    cell->is_macro = 0/*FALSE*/;
    cell->child = add_sequence(cell->child, s+1, s_len-1, expansion, exp_len); // add rest of word recursively, although a loop would work here
  }
  return base;
}

void load_database(struct NODE **newbase)
{
  int i, bp, sibidx, childidx;
  FILE *db = fopen("/tmp/testdb.trie", "r");
  if (db == NULL) {
    cellarray = malloc(1000000 * sizeof(struct NODE));
    if (cellarray == NULL) {
      fprintf(stderr, "vide: cannot malloc initial array\n");
      exit(1);
    }
    nextfree = 0;
    fprintf(stderr, "Will create new database /tmp/testdb.trie\n");
  } else {
    fscanf(db, "%d %d\n", &nextfree, &bp);
    fprintf(stderr, "Loading %d cells\n", nextfree);
    cellarray = malloc((nextfree+1000000) * sizeof(struct NODE));
    if (cellarray == NULL) {
      fprintf(stderr, "tokra: cannot malloc initial array\n");
      exit(1);
    }
    for (i = 0; i < nextfree; i++) {
      int j;
      fscanf(db, "%d ", &cellarray[i].c);
      fscanf(db, "%d ", &j); cellarray[i].ends_word = j;
      fscanf(db, "%d ", &j); cellarray[i].ends_sibling_list = j;
      fscanf(db, "%d ", &j); cellarray[i].is_macro = j;
      fscanf(db, "%d ", &cellarray[i].expansion_length);
      cellarray[i].expansion = malloc(sizeof(struct NODE *) * cellarray[i].expansion_length);
      for (j = 0; j < cellarray[i].expansion_length; j++) {
        fscanf(db, "%d ", &cellarray[i].expansion[j]);
      }
      fscanf(db, "%d ", &sibidx);
      if (sibidx == -1) {
        cellarray[i].sibling = NULL;
      } else {
        cellarray[i].sibling = (struct NODE *)((sibidx*sizeof(struct NODE) + (long long)&cellarray[0]));
      }
      fscanf(db, "%d\n", &childidx);
      if (childidx == -1) {
        cellarray[i].child = NULL;
      } else {
        cellarray[i].child = (struct NODE *)((childidx*sizeof(struct NODE) + (long long)&cellarray[0]));
      }
    }
    fprintf(stderr, "RELOADED base at %p\n", cellarray);
    // dump(cellarray, 0);  
  }
  *newbase = &cellarray[bp];
}

void dump_database(struct NODE *base)
{
  int i, bp = -1;
  FILE *db = fopen("/tmp/testdb.trie.new", "w");
  if (db == NULL) {
    fprintf(stderr, "Could not create new database /tmp/testdb.trie.new");
    exit(0);
  } else {
    fprintf(stderr, "Dumping %d cells to /tmp/testdb.trie.new\n", nextfree);
    for (i = 0; i < nextfree; i++) {
      if (base == &cellarray[i]) bp = i;
    }
    fprintf(db, "%d %d\n", nextfree, bp);

    // 3
    // 27 0 0 0 0 0 1
    // 79 0 1 0 0 -1 2
    // 117 1 1 1 8 97 98 99 100 101 102 103 104 -1 -1

    for (i = 0; i < nextfree; i++) {
      int j;
      fprintf(db, "%d ", cellarray[i].c);
      fprintf(db, "%d ", cellarray[i].ends_word);
      fprintf(db, "%d ", cellarray[i].ends_sibling_list);
      fprintf(db, "%d ", cellarray[i].is_macro);
      fprintf(db, "%d ", cellarray[i].expansion_length);
      for (j = 0; j < cellarray[i].expansion_length; j++) {
        fprintf(db, "%d ", cellarray[i].expansion[j]);
      }
      fprintf(db, "%lld ", (cellarray[i].sibling == NULL ? -1 : ((long long)cellarray[i].sibling - (long long)&cellarray[0])/sizeof(struct NODE)));
      fprintf(db, "%lld\n", (cellarray[i].child == NULL ? -1 : ((long long)cellarray[i].child - (long long)&cellarray[0])/sizeof(struct NODE)));
    }
  }
}
